package ueb04

import scala.util.parsing.combinator.{RegexParsers, Parsers}
import scala.language.postfixOps

sealed trait Expr
case class Var(n: String) extends Expr
case class Literal(n: Double) extends Expr
case class BinOp(op: String, l: Expr, r: Expr) extends Expr

sealed trait Stmt
case class Assign(l: Var, r: Expr) extends Stmt
case class Sequence(l: Stmt, r: Stmt) extends Stmt
case class Print(expr: Expr) extends Stmt
case class While(cond: Expr, body: Stmt) extends Stmt

object Expr {
  def binOp: PartialFunction[String, (Double, Double) => Double] = {
    case "+" => _ + _
    case "-" => _ - _
    case "*" => _ * _
    case "/" => _ / _
  }

  def calculate(expr: Expr): Double = expr match {
    case Literal(n)                             => n
    case BinOp(op,l,r) if binOp.isDefinedAt(op) => binOp(op)(calculate(l),calculate(r))
    case _ => Double.NaN
  }
}

object ExprParser extends RegexParsers {
  def block: Parser[Stmt] =
    statement ~ (";" ~> statement *)                 ^^ { case l~rs => rs.foldLeft(l)(Sequence) }

  def statement: Parser[Stmt] = (
    "PRINT" ~! expr                              ^^ { case _~e => Print(e) }
  | "WHILE" ~! (expr <~ "DO") ~ (block <~ "END") ^^ { case _~c~b => While(c,b)}
  | (variable <~ ":=") ~! expr                   ^^ { case v~e => Assign(v,e) }
  )

  /** combinator for left associative binary operators */
  def binopl(expr: Parser[Expr], ops: Parser[String]): Parser[Expr] =
    expr ~ ( ops ~ expr * ) ^^
      { case l ~ rs => rs.foldLeft(l) { case (l,op~r) => BinOp(op,l,r) } }

  def expr:        Parser[Expr] = binopl(term,"+"|"-")
  def term:        Parser[Expr] = binopl(factor,"*"|"/"|"%")

  def factor: Parser[Expr] = literal | variable | "(" ~> expr <~ ")"

  def variable: Parser[Var] = "[a-zA-Z][a-zA-Z0-9]*".r ^^ Var
  def literal: Parser[Expr] = "[0-9]+".r ^^ { s => Literal(s.toDouble) }
}